AGORA_KEYPOINTS = [
    'pelvis', 'left_hip', 'right_hip', 'spine_1', 'left_knee', 'right_knee',
    'spine_2', 'left_ankle', 'right_ankle', 'spine_3', 'left_foot',
    'right_foot', 'neck', 'left_collar', 'right_collar', 'head',
    'left_shoulder', 'right_shoulder', 'left_elbow', 'right_elbow',
    'left_wrist', 'right_wrist', 'jaw', 'left_eyeball', 'right_eyeball',
    'left_index_1', 'left_index_2', 'left_index_3', 'left_middle_1',
    'left_middle_2', 'left_middle_3', 'left_pinky_1', 'left_pinky_2',
    'left_pinky_3', 'left_ring_1', 'left_ring_2', 'left_ring_3',
    'left_thumb_1', 'left_thumb_2', 'left_thumb_3', 'right_index_1',
    'right_index_2', 'right_index_3', 'right_middle_1', 'right_middle_2',
    'right_middle_3', 'right_pinky_1', 'right_pinky_2', 'right_pinky_3',
    'right_ring_1', 'right_ring_2', 'right_ring_3', 'right_thumb_1',
    'right_thumb_2', 'right_thumb_3', 'nose', 'right_eye', 'left_eye',
    'right_ear', 'left_ear', 'left_bigtoe', 'left_smalltoe', 'left_heel',
    'right_bigtoe', 'right_smalltoe', 'right_heel', 'left_thumb', 'left_index',
    'left_middle', 'left_ring', 'left_pinky', 'right_thumb', 'right_index',
    'right_middle', 'right_ring', 'right_pinky', 'right_eyebrow_1',
    'right_eyebrow_2', 'right_eyebrow_3', 'right_eyebrow_4', 'right_eyebrow_5',
    'left_eyebrow_5', 'left_eyebrow_4', 'left_eyebrow_3', 'left_eyebrow_2',
    'left_eyebrow_1', 'nosebridge_1', 'nosebridge_2', 'nosebridge_3',
    'nosebridge_4', 'nose_1', 'nose_2', 'nose_3', 'nose_4', 'nose_5',
    'right_eye_1', 'right_eye_2', 'right_eye_3', 'right_eye_4', 'right_eye_5',
    'right_eye_6', 'left_eye_4', 'left_eye_3', 'left_eye_2', 'left_eye_1',
    'left_eye_6', 'left_eye_5', 'mouth_1', 'mouth_2', 'mouth_3', 'mouth_4',
    'mouth_5', 'mouth_6', 'mouth_7', 'mouth_8', 'mouth_9', 'mouth_10',
    'mouth_11', 'mouth_12', 'lip_1', 'lip_2', 'lip_3', 'lip_4', 'lip_5',
    'lip_6', 'lip_7', 'lip_8'
]
