HYBRIK_29_KEYPOINTS = [
    'pelvis',
    'left_hip',
    'right_hip',  # 2
    'spine_1',
    'left_knee',
    'right_knee',  # 5
    'spine_2',
    'left_ankle',
    'right_ankle',  # 8
    'spine_3',
    'left_foot',
    'right_foot',  # 11
    'neck',
    'left_collar',
    'right_collar',  # 14
    'jaw',  # 15
    'left_shoulder',
    'right_shoulder',  # 17
    'left_elbow',
    'right_elbow',  # 19
    'left_wrist',
    'right_wrist',  # 21
    'left_thumb',
    'right_thumb',  # 23
    'head',
    'left_middle',
    'right_middle',  # 26
    'left_bigtoe',
    'right_bigtoe'  # 28
]
