PW3D_KEYPOINTS = [
    'nose',
    'neck_extra',
    'right_shoulder',
    'right_elbow',
    'right_wrist',
    'left_shoulder',
    'left_elbow',
    'left_wrist',
    'right_hip_extra',
    'right_knee',
    'right_ankle',
    'left_hip_extra',
    'left_knee',
    'left_ankle',
    'right_eye',
    'left_eye',
    'right_ear',
    'left_ear',
]
