MPI_INF_3DHP_KEYPOINTS = [
    'spine_3',
    'spine_4_3dhp',
    'spine_2',
    'spine_extra',  # close to spine2
    'pelvis_extra',
    'neck_extra',  # throat
    'head_extra',
    'headtop',
    'left_clavicle_3dhp',
    'left_shoulder',
    'left_elbow',
    'left_wrist',
    'left_hand_3dhp',
    'right_clavicle_3dhp',
    'right_shoulder',
    'right_elbow',
    'right_wrist',
    'right_hand_3dhp',
    'left_hip_extra',
    'left_knee',
    'left_ankle',
    'left_foot',
    'left_toe_3dhp',
    'right_hip_extra',
    'right_knee',
    'right_ankle',
    'right_foot',
    'right_toe_3dhp'
]

MPI_INF_3DHP_TEST_KEYPOINTS = [
    'headtop',
    'neck_extra',
    'right_shoulder',
    'right_elbow',
    'right_wrist',
    'left_shoulder',
    'left_elbow',
    'left_wrist',
    'right_hip_extra',
    'right_knee',
    'right_ankle',
    'left_hip_extra',
    'left_knee',
    'left_ankle',
    'pelvis_extra',
    'spine_extra',  # close to spine2
    'head_extra'
]

HYBRIK_MPI_INF_3DHP_KEYPOINTS = [
    'spine_3',
    'spine_4_3dhp',
    'spine_2',
    'spine_extra',  # close to spine2
    'pelvis',
    'neck',  # throat
    'head_extra',
    'headtop',
    'left_clavicle_3dhp',
    'left_shoulder',
    'left_elbow',
    'left_wrist',
    'left_hand_3dhp',
    'right_clavicle_3dhp',
    'right_shoulder',
    'right_elbow',
    'right_wrist',
    'right_hand_3dhp',
    'left_hip',
    'left_knee',
    'left_ankle',
    'left_foot',
    'left_toe_3dhp',
    'right_hip',
    'right_knee',
    'right_ankle',
    'right_foot',
    'right_toe_3dhp'
]
